#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import json

from pyxie.parsing.context import Context
from pyxie.model.tree import Tree
from pyxie.model.functions import builtins

MULTI_TYPES_WARN = False
WARNINGS_ARE_FAILURES = False

# This lookup should probably really look somewhere else
expression_mixed_types = {
       #(function_tag, type, type ) -> result_type
        ("op_times", "integer", "string") : "string",
        ("op_times", "string","integer") : "string",
        ("op_times", "integer", "char") : "string",
        ("op_times", "char","integer") : "string",
    }

# Utility functions

def jdump(thing):
    "Calls __json__ on a thing to try and convert it into a json serialisable thing"
    try:
        return thing.__json__()
    except AttributeError:
        print "WARNING::", repr(thing), "is not a pynode"
        print "       ::", type(thing)

def warn(message):
    if WARNINGS_ARE_FAILURES:
        raise Exception(message)
    else:
        print message

# Astract base nodes

class PyNode(Tree):
    """Representation of a python node"""
    tag = "node"
    ntype = None # Type for this node
    def __init__(self, *args):
        # Initialise the tree
        super(PyNode,self).__init__()
    def tags(self):
        return ["node"]
    def classname(self):
        return self.__class__.__name__
    def __info__(self):
        return { self.tag : {"type":self.ntype} }

class PyOperation(PyNode):
    tag = "operation"
    def __init__(self, *args):
        super(PyOperation,self).__init__()

class PyStatement(PyNode):
    tag = "statement"
    def __init__(self, *args):
        super(PyStatement,self).__init__()

# Grammar nodes, in approximate grammar order

class PyProgram(PyNode):
    tag = "program"
    def __init__(self, statements):
        super(PyProgram,self).__init__()
        self.statements = statements
        self.includes = None
        self.add_children(statements)
    def __repr__(self):
        return "PyProgram(%s)" % (repr(self.statements), )
    def __json__(self):
        return [ self.tag, jdump(self.statements) ]

    def __info__(self):
        info = super(PyProgram, self).__info__()
        info[self.tag].update(self.statements.__info__())
        contexts = Context.contexts
        contexts_info = []
        for context in contexts:
            contexts_info.append(contexts[context].__json__())
        info[self.tag]["contexts"] = contexts_info
        return info

    def analyse(self):
        print "ANALYSING PROGRAM"

        global_context = Context()
        for node in self.depth_walk():
            if node.tag == "identifier":
                node.context = global_context
                print "NODE", node

        self.ntype = self.get_type()
        self.statements.analyse() # Descend through the tree

    def get_type(self):
        # Program has no value so no type
        return None

class PyBlock(PyNode):
    tag = "block"
    def __init__(self, statements):
        super(PyBlock,self).__init__()
        self.statements = statements
        self.add_children(statements)
    def __repr__(self):
        return "PyBlock(%s)" % (repr(self.statements), )
    def __json__(self):
        return [ self.tag, jdump(self.statements) ]

    def __info__(self):
        # Minimal change from "Program, since might be considered similar)
        # That said, I suspect that's wrong since a BLOCK doesn't necessarily
        # have a new context. Commenting that out for the moment.
        info = super(PyBlock, self).__info__()
        info[self.tag].update(self.statements.__info__())
#        contexts = Context.contexts
#        contexts_info = []
#        for context in contexts:
#            contexts_info.append(contexts[context].__json__())
#        info[self.tag]["contexts"] = contexts_info
        return info

    def analyse(self):
        print "ANALYSING BLOCK"

#        global_context = Context()
#        for node in self.depth_walk():
#            if node.tag == "identifier":
#                node.context = global_context

        self.ntype = None
        self.statements.analyse() # Descend through the tree

    def get_type(self):
        # Program has no value so no type
        return None

class PyStatements(PyNode):
    tag = "statements"
    def __init__(self, head, *tail):
        super(PyStatements,self).__init__()
        if not isinstance(head, PyEmptyStatement): #Filter out empty statements here.
            self.statements = [ head ]
        else:
            self.statements = [ ]
        if tail:
            for node in tail:
                self.statements = self.statements + node.statements
        if len(self.statements) > 0:
            self.add_children(*(self.statements))

    def __info__(self):
        info = super(PyStatements, self).__info__()
        info[self.tag]["block"] = [ x.__info__() for x in self.statements]
        return info

    def __repr__(self):
        return "PyStatements(%s)" % ",\n ".join([repr(x) for x in self.statements])

    def __json__(self):
        return [self.tag, [ jdump(x) for x in self.statements] ]

    def __iter__(self):
        for statement in self.statements:
            yield statement

    def analyse(self):
        print "ANALYSING STATEMENTS"
        self.ntype = self.get_type()
        for statement in self.statements:
            statement.analyse() # Descend through the tree

    def get_type(self):
        # Block of statements has no value, so no type
        return None

class PyAssignment(PyStatement):
    tag = "assignment_statement"
    def __init__(self, lvalue, rvalue, assign_type):
        super(PyAssignment,self).__init__()
        self.lvalue = lvalue
        self.rvalue = rvalue
        self.assign_type = assign_type
        self.add_children(self.lvalue, self.rvalue)

    def __repr__(self):
        return "PyAssignment(%s,%s,%s)" % (repr(self.lvalue), repr(self.rvalue), repr(self.assign_type))
    def __json__(self):
        return [ self.tag, jdump(self.lvalue), jdump(self.rvalue), self.assign_type ]

    def __info__(self):
        info = super(PyAssignment, self).__info__()
        info[self.tag]["lvalue"] = self.lvalue.__info__()
        info[self.tag]["rvalue"] = self.rvalue.__info__()
        info[self.tag]["assign_type"] = self.assign_type
        return info

    def analyse(self):
        print "ANALYSING ASSIGNMENT"
        print "ANALYSE RIGHT"
        self.rvalue.analyse()
        self.lvalue.add_rvalue(self.rvalue)
        self.lvalue.analyse()

        self.ntype = self.get_type()

    def get_type(self):
        ltype = self.lvalue.get_type()
        rtype = self.rvalue.get_type()

        print "Type for lvalue:", ltype
        print "Type for rvalue:", rtype
        print "Types match:", rtype==ltype

        # rtype wins because it's being used to set the left
        return rtype

class PyExpressionStatement(PyStatement):
    tag = "expression_statement"
    def __init__(self, value):
        super(PyExpressionStatement,self).__init__()
        self.value = value
        self.add_children(value)

    def __repr__(self):
        return "PyExpressionStatement(%s)" % (repr(self.value), )
    def __json__(self):
        return [ self.tag, jdump(self.value) ]
    def __info__(self):
        info = super(PyExpressionStatement, self).__info__()
        info[self.tag]["value"] = self.value.__info__()
        return info
    def analyse(self):
        print "ANALYSING EXPRESSION STATEMENT"
        self.value.analyse()
        self.ntype = self.get_type()
    def get_type(self):
        return self.value.get_type()

class PyFunctionCall(PyStatement):
    tag = "function_call"
    def __init__(self, identifier, expr_list):
        super(PyFunctionCall,self).__init__()
        self.identifier = identifier
        self.expr_list = expr_list
        self.add_children(expr_list)
        self.builtin = False
        self.isiterator = False
    def __repr__(self):
        return "PyFunctionCall(%s, %s)" % (repr(self.identifier), repr(self.expr_list), )
    def __json__(self):
        return [ self.tag, jdump(self.identifier), jdump(self.expr_list) ]
    def __info__(self):
        info = super(PyFunctionCall, self).__info__()
        info[self.tag]["name"] = self.identifier.__info__()
        info[self.tag]["args"] = [ x.__info__() for x in self.expr_list ]
        return info

    def analyse(self):
        # We'll need to decorate the function call with information from somewhere
        # For now though, we won't
        print "ANALYSING FUNCTION CALL"
        if self.identifier.value in builtins:
            print "FUNCTION", self.identifier.value, "FOUND IN BUILTINS"
            self.builtin = True
            self.ntype = self.get_type()
            print "WE CAN DO ANALYSIS FOR LIMITED BUILTINS"
            print " IS AN ITERATOR:", self.isiterator
            print "  VALUE(S) TYPE:", self.ntype
            return
        print "NOTE: WE DON'T ACTUALLY DO ANY ANALYSIS YET - GENERALLY"
        print "NOTE: ASIDE FROM SPECIAL CASING WE'LL JUST PASS THROUGH"
        return

    def get_type(self):
        # function calls have no default value, so for now we'll return None
        # This will be improved later on.
        if self.builtin:
            print "It's a built in function, we can get it's type in a bit..."
            meta = builtins[self.identifier.value]
            self.isiterator = meta.get("iterator", False)
            if self.isiterator:
                return meta.get("values_type", None)
            return meta.get("return_type", None)
        print "GETTING FUNCTION TYPE"
        print "NOTE: WE CAN'T ACTUALLY GET THE TYPE YET"
        print "NOTE: ASIDE FROM SPECIAL CASING WE'LL JUST CROSS FINGERS FOR NOW"
        return None

class PyForLoop(PyStatement):
    tag = "for_statement"
    def __init__(self, identifier, expression, block):
        super(PyForLoop,self).__init__()
        self.identifier= identifier
        self.expression = expression
        self.block = block
        self.add_children(identifier, expression, block)
    def __repr__(self):
        return "PyForLoop(%s, %s, %s)" % (repr(self.identifier), repr(self.expression), repr(self.block) )
    def __json__(self):
        return [ self.tag, jdump(self.identifier), jdump(self.expression), jdump(self.block) ]
    def __info__(self):
        info = super(PyForLoop, self).__info__()
        info[self.tag]["identifier"] = self.identifier.__info__()
        info[self.tag]["block"] = self.block.__info__()
        info[self.tag]["expression"] = self.expression.__info__()
        return info
    def analyse(self):
        # We'll need to decorate the function call with information from somewhere
        # For now though, we won't
        print "ANALYSING FOOR LOOP"
        print"                ... Analyse expression"
        expression = self.expression
        expression.analyse()
        print expression
        if expression.isiterator:
            print "Expression is an iterator"

        self.identifier.add_rvalue(expression) # The result of the expression becomes a repeated rvalue for the identifier, so this sorta makes sense

        print"                ... Update identifier"
        print"                ... Analyse identifier"
        self.identifier.analyse()

        print"                ... Analyse block"
        self.block.analyse()
        return

class PyWhileStatement(PyStatement):
    tag = "while_statement"
    def __init__(self, condition, block):
        super(PyWhileStatement,self).__init__()
        self.condition = condition
        self.block = block
        self.add_children(condition, block)
    def __repr__(self):
        return "PyWhileStatement(%s, %s)" % (repr(self.condition), repr(self.block) )
    def __json__(self):
        return [ self.tag, jdump(self.condition), jdump(self.block) ]
    def __info__(self):
        info = super(PyWhileStatement, self).__info__()
        info[self.tag]["condition"] = self.condition.__info__()
        # info[self.tag]["block"] = [ x.__info__() for x in self.expr_list ]
        info[self.tag]["block"] = self.block.__info__()
        return info
    def analyse(self):
        # We'll need to decorate the function call with information from somewhere
        # For now though, we won't
        print "ANALYSING WHILE BLOCK"
        print "analyse expression, and analyse block"
        self.condition.analyse()
        self.block.analyse()
        return
    def get_type(self):
        # function calls have no default value, so for now we'll return None
        # This will be improved later on.
        print "GETTING WHILE BLOCK TYPE - which should be None - for now"
        return None

class PyIfStatement(PyStatement):
    tag = "if_statement"
    def __init__(self, condition, block, else_clause=None):
        super(PyIfStatement,self).__init__()
        self.condition = condition
        self.block = block
        self.else_clause = else_clause
        self.add_children(condition, block)
    def __repr__(self):
        else_clause = ""
        if self.else_clause:
            else_clause = ", %s" % (repr(self.else_clause),)
        return "PyIfStatement(%s, %s%s)" % (repr(self.condition), repr(self.block), else_clause)
    def __json__(self):
        else_clause = ""
        if self.else_clause:
            else_clause = jdump(self.else_clause)
        return [ self.tag, jdump(self.condition), jdump(self.block), "", else_clause ] # Space left for elif chain...

    def __info__(self):
        info = super(PyIfStatement, self).__info__()
        info[self.tag]["condition"] = self.condition.__info__()
        info[self.tag]["block"] = self.block.__info__()
        if self.else_clause:
            info[self.tag]["else_clause"] = self.else_clause.__info__()
        return info
    def analyse(self):
        print "ANALYSING IF BLOCK"
        print "analyse expression, and analyse block"
        self.condition.analyse()
        self.block.analyse()
        if self.else_clause:
            self.else_clause.analyse()
        return
    def get_type(self):
        # function calls have no default value, so for now we'll return None
        # This will be improved later on.
        print "GETTING IF BLOCK TYPE - which should be None - for now"
        return None

class PyElIfClause(PyStatement):
    tag = "elif_clause"
    def __init__(self, condition, block, else_clause=None):
        super(PyElIfClause,self).__init__()
        self.condition = condition
        self.block = block
        self.else_clause = else_clause
        self.add_children(condition, block)
    def __repr__(self):
        else_clause = ""
        if self.else_clause:
            else_clause = ", %s" % (repr(self.else_clause),)
        return "PyElIfClause(%s, %s%s)" % (repr(self.condition), repr(self.block), else_clause)
    def __json__(self):
        else_clause = ""
        if self.else_clause:
            else_clause = jdump(self.else_clause)
        return [ self.tag, jdump(self.condition), jdump(self.block), "", else_clause ] # Space left for elif chain...

    def __info__(self):
        info = super(PyElIfClause, self).__info__()
        info[self.tag]["condition"] = self.condition.__info__()
        info[self.tag]["block"] = self.block.__info__()
        if self.else_clause:
            info[self.tag]["else_clause"] = self.else_clause.__info__()
        return info
    def analyse(self):
        print "ANALYSING ELIF BLOCK"
        print "analyse expression, and analyse block"
        self.condition.analyse()
        self.block.analyse()
        if self.else_clause:
            self.else_clause.analyse()
        return
    def get_type(self):
        # function calls have no default value, so for now we'll return None
        # This will be improved later on.
        print "GETTING IF BLOCK TYPE - which should be None - for now"
        return None

class PyElseClause(PyStatement):
    tag = "else_clause"
    def __init__(self, block):
        super(PyElseClause,self).__init__()
        self.block = block
        self.add_children(block)
    def __repr__(self):
        return "PyElseClause(%s)" % (repr(self.block), )
    def __json__(self):
        return [ self.tag, jdump(self.block) ]
    def __info__(self):
        info = super(PyElseClause, self).__info__()
        info[self.tag]["block"] = self.block.__info__()
        return info
    def analyse(self):
        print "ANALYSING ELSE CLAUSE"
        print "analyse expression, and analyse block"
        self.block.analyse()
        return
    def get_type(self):
        # function calls have no default value, so for now we'll return None
        # This will be improved later on.
        print "GETTING ELSE CLAUSE TYPE - which should be None - for now"
        return None

class PyEmptyStatement(PyStatement):
    tag = "empty_statement"
    def __init__(self):
        super(PyEmptyStatement,self).__init__()
    def analyse(self):
        pass
    def __json__(self):
        return [ self.tag ]

class PyBreakStatement(PyStatement):
    tag = "break_statement"
    def __init__(self):
        super(PyBreakStatement,self).__init__()
    def analyse(self):
        pass
    def __json__(self):
        return [ self.tag ]

class PyContinueStatement(PyStatement):
    tag = "continue_statement"
    def __init__(self):
        super(PyContinueStatement,self).__init__()
    def analyse(self):
        pass
    def __json__(self):
        return [ self.tag ]

class PyPrintStatement(PyStatement):
    tag = "print_statement"
    def __init__(self, expr_list):
        super(PyPrintStatement,self).__init__()
        self.expr_list = expr_list
        self.add_children(expr_list)

    def __repr__(self):
        return "PyPrintStatement(%s)" % (repr(self.expr_list), )
    def __json__(self):
        return [ self.tag, jdump(self.expr_list) ]
    def __info__(self):
        info = super(PyPrintStatement, self).__info__()
        info[self.tag]["args"] = [ x.__info__() for x in self.expr_list ]
        return info
    def analyse(self):
        print "ANALYSING PRINT STATEMENT"
        for expr in self.expr_list:
            expr.analyse() # Descend through the tree

        self.ntype = self.get_type()
    def get_type(self):
        # Print statement has no return value or default value
        return None

class PyExprList(PyNode):
    tag = "expression_list"
    def __init__(self, expr, *tail):
        super(PyExprList,self).__init__()
        self.expressions = [ expr ]
        if tail:
            for node in tail:
                self.expressions = self.expressions + node.expressions
        self.add_children(*(self.expressions))

    def __repr__(self):
        return "PyExprList(%s)" % ",\n ".join([repr(x) for x in self.expressions])

    def __json__(self):
        return [self.tag, [ jdump(x) for x in self.expressions] ]

    def __iter__(self):
        for expression in self.expressions:
            yield expression

    def __info__(self):
        raise Exception("Expression List should not have info called directly")

    def analyse(self):
        raise Exception("Expression List should not have analyse called directly")

# Base node for expressions, and all operators
class PyOperator(PyOperation):
    tag = "operator"
    def __init__(self, arg1, arg2):
        super(PyOperator,self).__init__()
        self.arg1 = arg1
        self.arg2 = arg2
        self.ntype = None
        self.add_children(arg1,arg2)

    def __info__(self):
        info = super(PyOperator, self).__info__()
        info[self.tag]["arg1"] = self.arg1.__info__()
        info[self.tag]["arg2"] = self.arg2.__info__()
        return info

    @property
    def type(self):
        if self.ntype != None:
            return self.ntype
        try:
            if self.arg1.get_type() == self.arg2.get_type():
                self.ntype = self.arg1.get_type()
            elif (self.tag, self.arg1.get_type(),self.arg2.get_type()) in expression_mixed_types:
                self.ntype = expression_mixed_types[self.tag, self.arg1.get_type(),self.arg2.get_type()]
            else:
                self.ntype = "Mixed types, need to resolve", self.tag, self.arg1.get_type(),self.arg2.get_type()
            return self.ntype
        except:
            print "TAG", self.tag
            print "ARG1", self.arg1.get_type()
            print "ARG2", self.arg2.get_type()
            raise

    def __repr__(self):
        return "%s(%s, %s)" % (self.classname(),repr(self.arg1),repr(self.arg2))
    def __json__(self):
        return [ self.tag, jdump(self.arg1), jdump(self.arg2) ]
    def get_type(self):
        return self.type
    def analyse(self):
        print "ANALYSING OPERATOR", self.tag
        self.arg1.analyse()
        self.arg2.analyse()

        self.ntype = self.get_type()

class PyBoolOperator(PyOperation):
    pass

class PyOrOperator(PyBoolOperator):
    tag = "or_operator"
    def __init__(self, arg1, arg2):
        super(PyOrOperator,self).__init__()
        self.arg1 = arg1
        self.arg2 = arg2
        self.ntype = None
        self.add_children(arg1,arg2)

    def __info__(self):
        info = super(PyOrOperator, self).__info__()
        info[self.tag]["arg1"] = self.arg1.__info__()
        info[self.tag]["arg2"] = self.arg2.__info__()
        return info

    @property
    def type(self):
        return "bool"

    def __repr__(self):
        return "%s(%s, %s)" % (self.classname(),repr(self.arg1),repr(self.arg2))
    def __json__(self):
        return [ self.tag, jdump(self.arg1), jdump(self.arg2) ]
    def get_type(self):
        return "bool"
    def analyse(self):
        print "ANALYSING OPERATOR:", self.tag
        self.arg1.analyse()
        self.arg2.analyse()

        self.ntype = self.get_type()

class PyAndOperator(PyBoolOperator):
    tag = "and_operator"
    def __init__(self, arg1, arg2):
        super(PyAndOperator,self).__init__()
        self.arg1 = arg1
        self.arg2 = arg2
        self.ntype = None
        self.add_children(arg1,arg2)

    def __info__(self):
        info = super(PyAndOperator, self).__info__()
        info[self.tag]["arg1"] = self.arg1.__info__()
        info[self.tag]["arg2"] = self.arg2.__info__()
        return info

    @property
    def type(self):
        return "bool"

    def __repr__(self):
        return "%s(%s, %s)" % (self.classname(),repr(self.arg1),repr(self.arg2))
    def __json__(self):
        return [ self.tag, jdump(self.arg1), jdump(self.arg2) ]
    def get_type(self):
        return "bool"
    def analyse(self):
        print "ANALYSING OPERATOR:", self.tag
        self.arg1.analyse()
        self.arg2.analyse()

        self.ntype = self.get_type()

class PyNotOperator(PyBoolOperator):
    tag = "not_operator"
    def __init__(self, arg1):
        super(PyNotOperator,self).__init__()
        self.arg1 = arg1
        self.ntype = None
        self.add_children(arg1)

    def __info__(self):
        info = super(PyNotOperator, self).__info__()
        info[self.tag]["arg1"] = self.arg1.__info__()
        return info

    @property
    def type(self):
        return "bool"

    def __repr__(self):
        return "%s(%s)" % (self.classname(),repr(self.arg1),)
    def __json__(self):
        return [ self.tag, jdump(self.arg1) ]
    def get_type(self):
        return "bool"
    def analyse(self):
        print "ANALYSING OPERATOR:", self.tag
        self.arg1.analyse()
        self.ntype = self.get_type()

class PyTimesOperator(PyOperator):
    tag = "op_times"

class PyDivideOperator(PyOperator):
    tag = "op_divide"

class PyPowerOperator(PyOperator):
    tag = "op_power"

class PyPlusOperator(PyOperator):
    tag = "op_plus"

class PyComparisonOperator(PyOperation):
    tag = "comparison_operator"
    def __init__(self, comparison, arg1, arg2):
        super(PyComparisonOperator,self).__init__()
        self.comparison = comparison
        self.arg1 = arg1
        self.arg2 = arg2
        self.ntype = None
        self.add_children(arg1,arg2)

        print "CREATED COMPARISON OPERATOR", comparison, arg1, arg2

    def __info__(self):
        info = super(PyComparisonOperator, self).__info__()
        info[self.tag]["comparison"] = self.comparison
        info[self.tag]["arg1"] = self.arg1.__info__()
        info[self.tag]["arg2"] = self.arg2.__info__()
        return info

    @property
    def type(self):
        return "bool"

    def __repr__(self):
        return "%s(%s, %s, %s)" % (self.classname(),self.comparison, repr(self.arg1),repr(self.arg2))
    def __json__(self):
        return [ self.tag, self.comparison, jdump(self.arg1), jdump(self.arg2) ]
    def get_type(self):
        return "bool"
    def analyse(self):
        print "ANALYSING", self.tag
        self.arg1.analyse()
        self.arg2.analyse()

        self.ntype = self.get_type()

class PyMinusOperator(PyOperator):
    tag = "op_minus"

# Base class for all Value Literals
class PyValueLiteral(PyNode):
    tag = "value_literal"
    def __init__(self, lineno, value):
        super(PyValueLiteral,self).__init__()
        self.lineno = lineno
        self.value = value
    def __repr__(self):
        return "%s(%d, %s)" % (self.classname(),self.lineno, repr(self.value))
    def __json__(self):
        return [ self.tag, self.lineno, self.value ]
    def __info__(self):
        info = super(PyValueLiteral, self).__info__()
        info[self.tag]["lineno"] = self.lineno
        info[self.tag]["value"] = self.value
        return info
    def analyse(self):
        print "ANALYSING VALUE LITERAL", self.tag
        # Don't go into containers, because there aren't any
        self.ntype = self.get_type()
    def get_type(self):
        raise NotImplementedError("PyValueLiteral does not have any implicit type - its subtypes do")

# All non-number value literals first
class PyString(PyValueLiteral):
    tag = "string"
    def get_type(self):
        return "string"

class PyCharacter(PyValueLiteral):
    tag = "character"
    def get_type(self):
        return "char"

class PyBoolean(PyValueLiteral):
    tag = "boolean"
    def get_type(self):
        return "bool"

# Resist the urge to put PyIdentifiers into a LUT immediately.
class PyIdentifier(PyValueLiteral):
    tag = "identifier"
    def __init__(self, *args):
        super(PyIdentifier, self).__init__(*args)
        self.context = None
        self.types = []

    def add_rvalue(self, expression):
        self.context.store(self.value, expression)

    def __info__(self):
        info = super(PyIdentifier, self).__info__()
        info[self.tag]["context "] = self.context
        info[self.tag]["types"] = self.types
        return info

    def get_type(self):
        return self.ntype

    def analyse(self):
        expression = self.context.lookup(self.value)
        self.ntype = expression.get_type()

# Base class for all numbers
class PyNumber(PyValueLiteral):
    tag = "number"
    def negate(self):
        self.value = - self.value
        return self

class PyFloat(PyNumber):
    tag = "float"
    def get_type(self):
        return "float"

class PyInteger(PyNumber):
    tag = "integer"
    def get_type(self):
        return "integer"

class PyHex(PyNumber):
    tag = "hex"
    def get_type(self):
        return "integer"

class PyOctal(PyNumber):
    tag = "octal"
    def get_type(self):
        return "integer"

class PyBinary(PyNumber):
    tag = "binary"
    def get_type(self):
        return "integer"

if __name__ == "__main__":
    trees = [
                PyMinusOperator(PyInteger(1,10),PyInteger(1,10)),
                PyBoolean(1,True),
                PyAssignment(PyIdentifier(1,"hello"), PyString(1,"world"), "="),
                PyExpressionStatement(PyMinusOperator(PyInteger(1,10),PyInteger(1,10)))
        ]
    for tree in trees:
        print tree

    for tree in trees:
        print jdump(tree)

    MULTI_TYPES_WARN = True
    WARNINGS_ARE_FAILURES = False

    ident = PyIdentifier(1,"hello")
    print ident, ident.get_type()

    ident.add_type("string")
    print ident, ident.get_type()
    ident.add_type("string")
    print ident, ident.get_type()
    ident.add_type("char")
    print ident, ident.get_type()
    ident.add_type("integer")
    print ident, ident.get_type()
    ident.add_type("float")
    print ident, ident.get_type()
    ident.add_type("bool")
    print ident, ident.get_type()

    trees = [
                PyString(1,"world"),
                PyBoolean(1,True),
                PyFloat(1,1.1),
                PyMinusOperator(PyInteger(1,10),PyInteger(1,10)),
                PyMinusOperator(PyFloat(1,10),PyFloat(1,10)),
                PyTimesOperator(PyInteger(1,10),PyString(1,"Hello")),
                PyTimesOperator(PyString(1,"Hello"),PyInteger(1,10)),
                PyTimesOperator(PyInteger(1,10),PyCharacter(1,"Hello")),
                PyTimesOperator(PyCharacter(1,"Hello"),PyInteger(1,10)),
                PyTimesOperator(PyInteger(1,10),PyBoolean(1,True)),
                PyTimesOperator(PyCharacter(1,"Hello"),PyFloat(1,1.1)),
                PyPlusOperator(PyInteger(1,10),PyString(1,"Hello")),

            ]

    for tree in trees:
        print tree, tree.get_type()

    #print PyMinusOperator(PyInteger(1,10),PyInteger(1,10))
    #print jdump(PyMinusOperator(PyInteger(1,10),PyInteger(1,10)))

    #print PyBoolean(1,True)
    #print jdump(PyBoolean(1,True))

    #print PyAssignment(PyIdentifier(1,"hello"), PyString(1,"world"), "=")

    #print jdump(PyAssignment(PyIdentifier(1,"hello"), PyString(1,"world"), "="))





#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ply.yacc as yacc

from pyxie.model.pynode import *
from pyxie.parsing.lexer import tokens
from pyxie.parsing.context import *

class Grammar(object):
    precedence = (
        ('left', 'PLUS','MINUS'),
        ('left', 'TIMES','DIVIDE'),
        ('right', 'UMINUS')
    )
    tokens = tokens

    def p_error(self,p):
        print "Syntax error at", p

    def p_program(self, p):
        "program : statements"
        p[0] = PyProgram(p[1])

    def p_statements_1(self, p):
        "statements : statement"
        p[0] = PyStatements(p[1])

    def p_statements_2(self, p):
        "statements : statement statements"
        p[0] = PyStatements(p[1], p[2])

    def p_statement_1(self, p):
        "statement : assignment_statement"
        p[0] = p[1]

    def p_statement_2(self, p):
        "statement : print_statement"
        p[0] = p[1]

    def p_statement_3(self, p):
        "statement : general_expression"
        p[0] = PyExpressionStatement(p[1])

    def p_statement_4_empty(self, p):
        "statement : EOL"
        p[0] = PyEmptyStatement()

    def p_statement_5(self, p):
        "statement : while_statement"
        p[0] = p[1]

    def p_statement_6(self, p):
        "statement : break_statement"
        p[0] = p[1]

    def p_statement_7(self, p):
        "statement : continue_statement"
        p[0] = p[1]

    def p_statement_8(self, p):
        "statement : if_statement"
        p[0] = p[1]

    def p_statement_9(self, p):
        "statement : for_statement"
        p[0] = p[1]

    def p_break_statement_1(self, p):
        "break_statement : BREAK"
        p[0] = PyBreakStatement()

    def p_continue_statement_1(self, p):
        "continue_statement : CONTINUE"
        p[0] = PyContinueStatement()

    def p_print_statement_1(self, p):
        "print_statement : PRINT expr_list"
        p[0] = PyPrintStatement(p[2])


    def p_while_statement_1(self, p):
        "while_statement : WHILE general_expression COLON EOL statement_block"
        p[0] = PyWhileStatement(p[2], p[5]) # pass in expression and block
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED WHILE"
        print "------------------> COND  :", p[2]
        print "------------------> BLOCK :", p[5]
        print "------------------> RESULT :", p[0]

    def p_if_statement_1(self, p):
        "if_statement : IF general_expression COLON EOL statement_block"
        print "PARSES IF"
        p[0] = PyIfStatement(p[2], p[5]) # pass in expression and block
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED IF STATEMENT"
        print "------------------> COND  :", p[2]
        print "------------------> BLOCK :", p[5]
        print "------------------> RESULT :", p[0]

    def p_if_statement_2(self, p):
        "if_statement : IF general_expression COLON EOL statement_block extended_if_clauses"
        print "PARSES IF"
        p[0] = PyIfStatement(p[2], p[5],else_clause=p[6]) # pass in expression and block
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED IF STATEMENT"
        print "------------------> COND  :", p[2]
        print "------------------> BLOCK :", p[5]
        print "------------------> RESULT :", p[0]

    def p_extended_if_clauses_1(self,p):
        "extended_if_clauses : else_clause"
        p[0] = p[1]

    def p_extended_if_clauses_2(self,p):
        "extended_if_clauses : elif_clause"
        p[0] = p[1]

    def p_else_clause_1(self,p):
        "else_clause : ELSE COLON EOL statement_block "
        p[0] = PyElseClause(p[4])


    def p_elif_clause_1(self, p):
        "elif_clause : ELIF general_expression COLON EOL statement_block"
        print "PARSES IF"
        p[0] = PyElIfClause(p[2], p[5]) # pass in expression and block
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED IF STATEMENT"
        print "------------------> COND  :", p[2]
        print "------------------> BLOCK :", p[5]
        print "------------------> RESULT :", p[0]

    def p_elif_clause_2(self, p):
        "elif_clause : ELIF general_expression COLON EOL statement_block extended_if_clauses"
        print "PARSES IF"
        p[0] = PyElIfClause(p[2], p[5],else_clause=p[6]) # pass in expression and block
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED IF STATEMENT"
        print "------------------> COND  :", p[2]
        print "------------------> BLOCK :", p[5]
        print "------------------> RESULT :", p[0]


    def p_for_statement_1(self, p):
        "for_statement : FOR IDENTIFIER IN general_expression COLON EOL statement_block"
        identifier = PyIdentifier(p.lineno(1), p[2])
        expression = p[4]
        statement_block = p[7]
        p[0] = PyForLoop(identifier, expression, statement_block)

    def p_block_1(self, p):
        "statement_block : INDENT statements DEDENT"
        p[0] = PyBlock(p[2])
        print
        print "SO WE REACHED BLOCK"
        print "------------------> STATEMENTS :", p[2]
        print "------------------> RESULT :", p[0]

    def p_expr_list_1(self,p):
        "expr_list : general_expression"
        p[0] = PyExprList(p[1])

    def p_expr_list_2(self,p):
        "expr_list : general_expression COMMA expr_list"
        p[0] = PyExprList(p[1],p[3])


    def p_assignment_statement(self, p):
        "assignment_statement : IDENTIFIER ASSIGN general_expression"
        identifier = PyIdentifier(p.lineno(1), p[1])

        p[0] = PyAssignment(identifier, p[3], p[2])

    def p_general_expression_1(self, p):
        "general_expression : boolean_expression"
        p[0] = p[1]

    def p_boolean_expression_1(self, p):
        "boolean_expression : boolean_and_expression"
        p[0] = p[1]

    def p_boolean_expression_2(self, p):
        "boolean_expression : boolean_expression OR boolean_and_expression"
        p[0] = PyOrOperator(p[1],p[3])

        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED AN OR EXPRESSION"
        print "------------------> LARG:", p[1]
        print "------------------> RARG :", p[3]
        print "------------------> RESULT :", p[0]

    def p_boolean_and_expression_1(self, p):
        "boolean_and_expression : boolean_not_expression"
        p[0] = p[1]

    def p_boolean_and_expression_2(self, p):
        "boolean_and_expression  : boolean_and_expression AND boolean_not_expression"
        p[0] = PyAndOperator(p[1],p[3])
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED AN AND EXPRESSION"
        print "------------------> LARG:", p[1]
        print "------------------> RARG :", p[3]
        print "------------------> RESULT :", p[0]

    def p_boolean_not_expression_1(self, p):
        "boolean_not_expression : relational_expression"
        p[0] = p[1]

    def p_boolean_not_expression_2(self, p):
        "boolean_not_expression : NOT boolean_not_expression "
        p[0] = PyNotOperator(p[2])
        print
        print "WARNING, expression used in a location requiring truthiness"                                     # TODO
        print "This will generally be OK for bools and integers but need a function for anything else"          # TODO
        print "In particular for strings, lists, dictionaries, tuples and so on"                                # TODO
        print "That is a separate card in the backlog though"                                                   # TODO
        print "SO WE REACHED AN AND EXPRESSION"
        print "------------------> ARG:", p[2]
        print "------------------> RESULT :", p[0]

    def p_relational_expression_1(self, p):
        "relational_expression : relational_expression COMPARISON_OPERATOR expression"
        p[0] = PyComparisonOperator(p[2], p[1], p[3])

    def p_relational_expression_2(self, p):
        "relational_expression : expression"
        p[0] = p[1]

    def p_expression_1(self, p):
        "expression : arith_expression"
        p[0] = p[1]

    def p_expression_2(self, p):
        "expression : expression PLUS arith_expression"
        p[0] = PyPlusOperator(p[1],p[3])

    def p_expression_3(self, p):
        "expression : expression MINUS arith_expression"
        p[0] = PyMinusOperator(p[1],p[3])

    def p_expression_4(self, p):
        "expression : expression POWER arith_expression"
        p[0] = PyPowerOperator(p[1],p[3])

    def p_arith_expression_1(self, p):
        "arith_expression     : expression_atom"
        p[0] = p[1]

    def p_arith_expression_2(self, p):
        "arith_expression     : arith_expression TIMES expression_atom"
        p[0] = PyTimesOperator(p[1],p[3])

    def p_arith_expression_3(self, p):
        "arith_expression     : arith_expression DIVIDE expression_atom"
        p[0] = PyDivideOperator(p[1],p[3])

    def p_expression_atom_1(self, p):
        "expression_atom : value_literal"
        p[0] = p[1]

    def p_expression_atom_2(self, p):
        "expression_atom : IDENTIFIER PARENL expr_list PARENR"
        p[0] = PyFunctionCall(PyIdentifier(p.lineno(1), p[1]), p[3])

    def p_expression_atom_3(self, p):
        "expression_atom : PARENL general_expression PARENR"
        p[0] = p[2]

    ### Core Literals

    def p_value_literal_0(self, p):
        "value_literal : number"
        p[0] = p[1]

    def p_value_literal_1(self, p):
        "number : NUMBER"
        p[0] = PyInteger(p.lineno(1), p[1])

    def p_value_literal_2(self, p):
        "number : FLOAT"
        p[0] = PyFloat(p.lineno(1), p[1])

    def p_value_literal_3(self, p):
        "number : HEX"
        p[0] = PyHex(p.lineno(1), p[1])

    def p_value_literal_4(self, p):
        "number : OCTAL"
        p[0] = PyOctal(p.lineno(1), p[1])

    def p_value_literal_5(self, p):
        "number : BINARY"
        p[0] = PyBinary(p.lineno(1), p[1])

    def p_value_literal_5a(self,p):
        "number : MINUS number %prec UMINUS"
        p[0] = p[2].negate()

    def p_value_literal_6(self, p):
        "value_literal : STRING"
        p[0] = PyString(p.lineno(1), p[1])

    def p_value_literal_6a(self, p):
        "value_literal : CHARACTER"
        p[0] = PyCharacter(p.lineno(1), p[1])

    def p_value_literal_7(self, p):
        "value_literal : BOOLEAN"
        p[0] = PyBoolean(p.lineno(1), p[1])

    def p_value_literal_8(self, p):
        "value_literal : IDENTIFIER"
        p[0] = PyIdentifier(p.lineno(1), p[1])

def parse(source,lexer):

   if not source.endswith("\n"): # Be relaxed about end of file
       source = source+ "\n"

   yacc.yacc(module=Grammar())
   result = yacc.parse(source, lexer=lexer)
   return result



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# This is currently working from the AST as represented by lists,
# switching to intermediate lists and then to nodes.

from pyxie.model.pynode import jdump
import pyxie.model.pynode as nodes
from pyxie.model.functions import builtins

iterator_unique_base = 0

def todo(*args):
    print "TODO", " ".join([repr(x) for x in args])

# Assumes that the analysis phase has taken place
def find_variables(AST):
    global iterator_unique_base
    variables = {}
    for node in AST.depth_walk():
        if node.tag == "for_statement":
            lvalue = node.identifier
            rvalue_source = node.expression
            identifier = lvalue.value
            v_type = lvalue.ntype
            print "FOR LOOP, IDENTIFIER", identifier, "HAS TYPE", v_type
            variables[identifier] = v_type
            if node.expression.isiterator:
                print "#####################################################"
                print "Extracting generator intermediate variable"
                print node.expression
                if isinstance(node.expression, nodes.PyFunctionCall):
                    identifier = node.expression.identifier
                    print "Bibble", identifier
                    print "Bibble", identifier.value
                    iterator_type = identifier.value
                    iterator_unique_base += 1
                    iterator_name = identifier.value + "_iter_" + str(iterator_unique_base)
                    print "iterator_type", iterator_type
                    print "iterator_name", iterator_name
                    variables[iterator_name] = iterator_type
                    node.expression.ivalue_name = iterator_name

                print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"

        if node.tag == "assignment_statement":
            if node.assign_type != "=":
                todo("find_variables - assignment where the assign_type is not '='")
                continue # Skip
            if node.lvalue.tag != "identifier":
                todo("find_variables - assignment where the lvalue is not an identifier")
                continue # Skip

            lvalue, rvalue, assign_type = node.lvalue, node.rvalue, node.assign_type
            identifier = lvalue.value
            v_type = lvalue.ntype

            if identifier in variables:
                todo("we could check that the identifier doesn't change type")
                continue # Skip
            variables[identifier] = v_type

    return variables

class UnknownType(Exception):
    pass

class CannotConvert(Exception):
    pass

def python_type_to_c_type(ptype):
    if ptype == "string":  return "string"
    if ptype == "integer":  return "int"
    if ptype == "bool": return "bool"
    if ptype == "float":   return "double"
    if ptype == "char": return "char"

    if ptype in builtins: return ptype

    raise UnknownType("Cannot identify C Type for %s" % ptype)

def includes_for_ctype(ctype):
    if ctype == "string":  return "<string>"

def includes_for_cstatement(cstatement):
    if cstatement[0] == "print_statement": return "<iostream>"

def crepr_literal(pyliteral):
    assert isinstance(pyliteral, nodes.PyValueLiteral)

    ptype = pyliteral.get_type()
    ctype = python_type_to_c_type(ptype)

    if ctype == "string":
        result = pyliteral.value
        result = result.replace('"','\\"')
        return '"' + result + '"'

    if ctype == "char":
        char = pyliteral.value
        char = char.replace("'","\\'")
        return "'" + char + "'"

    if ctype == "int":
        return repr(pyliteral.value)

    if ctype == "double":
        return repr(pyliteral.value)

    if ctype == "bool":
        if pyliteral.value:
            return "true"
        else:
            return "false"

    raise ValueError("Do not not know how to create crepr_literal for " + repr(pyliteral))

def crepr_op(py_op):
    assert ( isinstance(py_op, nodes.PyOperator) or isinstance(py_op, nodes.PyBoolOperator))
#    assert py_op[0] == "operator_function"
    func = py_op.tag

    if func == "op_plus":
        return ["op", "plus"]
    if func == "op_minus":
        return ["op", "minus"]
    if func == "op_times":
        return ["op", "times"]
    if func == "op_divide":
        return ["op", "divide"]
    if func == "or_operator":
        return ["op", "boolean_or"]
    if func == "and_operator":
        return ["op", "boolean_and"]
    if func == "not_operator":
        return ["op", "boolean_not"]
    else:
        todo("Cannot yet convert operators functions other than plus...")
        raise CannotConvert("Cannot yet convert operators functions other than plus...:" + repr(py_op))

def convert_assignment(assignment):
    lvalue, assign_type, rvalue = assignment.lvalue,assignment.assign_type, assignment.rvalue

    if assign_type != "=":
        todo("Convert Assignment where assign_type is not '='")
        raise CannotConvert("Cannot convert assignment where assign_type is not '='")

    if lvalue.tag != "identifier":
        todo("assignment where the lvalue is not an identifier")
        raise CannotConvert("Cannot convert assignment where the lvalue is not an identifier")

    if not ( isinstance(assignment.rvalue, nodes.PyOperator) or
             isinstance(assignment.rvalue, nodes.PyValueLiteral) or
             isinstance(assignment.rvalue, nodes.PyComparisonOperator) or
             isinstance(assignment.rvalue, nodes.PyBoolOperator)) :

        todo("assignment where the rvalue is not a value_literal or operator")
        raise CannotConvert("Cannot convert assignment where the rvalue is not a value_literal")

    # print rvalue
    clvalue = lvalue.value # FIXME: This is only valid for identifiers
    if isinstance(assignment.rvalue, nodes.PyValueLiteral):
        crvalue = crepr_literal(rvalue)

    if isinstance(assignment.rvalue, nodes.PyOperator):
        crvalue = convert_operator_function(rvalue)

    if isinstance(assignment.rvalue, nodes.PyBoolOperator):
        crvalue = convert_bool_operator_function(rvalue)

    if isinstance(assignment.rvalue, nodes.PyComparisonOperator):
        crvalue = convert_comparison(rvalue)

    return ["assignment", clvalue, "=", crvalue ]

def convert_value_literal(arg):
    # print repr(arg), arg
    stype = None
    tag, value, vtype, line = arg.tag, arg.value,arg.get_type(), arg.lineno
    if tag == "identifier":
        return ["identifier",  value]

    if vtype == "string":
        return ["string",  value]
    if vtype == "integer":
        return ["integer",  value]
    if vtype == "float":
        return ["double",  value]
    if vtype == "bool":
        if value == True:
            value = "true"
        else:
            value = "false"
        return ["boolean",  value]

    todo("CONVERSION: Cannot handle other value literals %s" % repr(arg))
    todo("CONVERSION: %s %s %s %d" % (tag, repr(value), repr(vtype), line))
    raise CannotConvert("Cannot convert value-literal of type" + repr(arg))



def convert_bool_operator_function(opfunc):
    print "CONVERT - convert_bool_operator_function", repr(opfunc)
    assert isinstance(opfunc, nodes.PyBoolOperator)

    func = opfunc.tag
    arg1 = opfunc.arg1

    if not isinstance(opfunc, nodes.PyNotOperator):
        arg2 = opfunc.arg2
    else:
        arg2 = None

    crepr_arg1 = convert_arg(arg1)
    if arg2:
        crepr_arg2 = convert_arg(arg2)
    else:
        crepr_arg2 = None

    print "crepr_arg1", repr(crepr_arg1)

    if arg2:
        print "crepr_arg2", repr(crepr_arg2)
        result = crepr_op(opfunc) + [crepr_arg1, crepr_arg2]
    else:
        result = crepr_op(opfunc) + [crepr_arg1 ]
    print repr(result)
    return result


def convert_operator_function(opfunc):
    print "CONVERT - convert_operator_function", repr(opfunc)
    assert isinstance(opfunc, nodes.PyOperator)

    func = opfunc.tag
    arg1 = opfunc.arg1
    arg2 = opfunc.arg2

    crepr_arg1 = convert_arg(arg1)
    crepr_arg2 = convert_arg(arg2)
    print "crepr_arg1", repr(crepr_arg1)
    print "crepr_arg2", repr(crepr_arg2)

    result = crepr_op(opfunc) + [crepr_arg1, crepr_arg2]
    print repr(result)
    return result

    #todo("Cannot yet convert operator functions")
    #raise CannotConvert("Cannot convert operator function :" + repr(arg))

def convert_comparison_operator(comparison):
    # t_NORMAL_COMPARISON_OPERATOR = r'(in|not +in|is|is +not)'

    if comparison in ["<", ">", "==", ">=", "<=", "<>", "!="]:
        return comparison

    raise NotImplementedError(repr(comparison))

def convert_comparison(comparison_spec):
    print "CONVERT - convert_comparison", repr(comparison_spec)
    assert isinstance(comparison_spec, nodes.PyComparisonOperator)

    comparison = comparison_spec.comparison
    arg1 = comparison_spec.arg1
    arg2 = comparison_spec.arg2

    crepr_comparison = convert_comparison_operator(comparison)
    crepr_arg1 = convert_arg(arg1)
    crepr_arg2 = convert_arg(arg2)
    print "crepr_arg1", repr(crepr_arg1)
    print "crepr_arg2", repr(crepr_arg2)

    result = ["op", crepr_comparison, crepr_arg1, crepr_arg2]
    print repr(result)
    return result



def convert_arg(arg):
    if isinstance(arg, nodes.PyValueLiteral):
        print "CONVERTING LITERAL", arg
        return convert_value_literal(arg)
    elif isinstance(arg, nodes.PyComparisonOperator):
        return convert_comparison(arg)
    elif isinstance(arg, nodes.PyOperator):
        return convert_operator_function(arg)
    elif isinstance(arg, nodes.PyBoolOperator):
        return  convert_bool_operator_function(arg)

    elif isinstance(arg, nodes.PyFunctionCall):
        print "NEED TO CONVERT FUNCTION CALL TO SOMETHING THE C CODE GENERATOR CAN HANDLE"
        cargs = []
        for expr in arg.expr_list:
            #print arg
            #print "We need to convert the arg", arg
            crepr = convert_arg(expr)
            carg = crepr
            cargs.append(carg)

        return ["function_call", convert_value_literal(arg.identifier),  cargs ]
        return ["Nothing"]
    else:
        todo("Handle print for non-value-literals")
        raise CannotConvert("Cannot convert print for non-value-literals")

def convert_print(print_statement):
    cstatement = []
    cargs = []
    for arg in print_statement.expr_list:
        #print arg
        #print "We need to convert the arg", arg
        crepr = convert_arg(arg)
        carg = crepr
        cargs.append(carg)
    return ["print_statement"] + cargs

def convert_while_statement(while_statement):
    crepr_condition = convert_arg(while_statement.condition)
    cstatements = convert_statements(while_statement.block)
    return ["while_statement", crepr_condition] + cstatements

import pprint
def convert_for_statement(for_statement):

    lvalue = for_statement.identifier
    rvalue_source = for_statement.expression
    step = for_statement.block

    clvalue = lvalue.value # FIXME: This is only valid for identifiers
    crvalue_source  = [ "iterator", convert_arg(rvalue_source) ]
    cstep = convert_statements(step)


    print "*******************"
    print crvalue_source 
    print "*******************"
    print "FOR STATEMENT :", 
    print "              : ", for_statement
    print "              :", dir(for_statement)
    print "     loop var :", for_statement.identifier
    print "loop var type :", for_statement.identifier.get_type()
    print "loop var ctype:", python_type_to_c_type(for_statement.identifier.get_type())
    print "     iterator :", for_statement.expression
    print "        block :", for_statement.block
    print "        info :", for_statement.__info__()
    print "*******************"
    pprint.pprint( for_statement.__info__() )
    print "*******************"
    # crepr_condition = convert_arg(for_statement.condition)
    # cstatements = convert_statements(while_statement.block)
#    return ["while_statement", crepr_condition] + cstatements
    return ["for_statement", clvalue, crvalue_source, cstep, for_statement ]


def convert_extended_clause(extended_clause):
    if extended_clause.tag == "elif_clause":
        print "WORKING THROUGH ELIF:", extended_clause
        crepr_condition = convert_arg(extended_clause.condition)
        cstatements = convert_statements(extended_clause.block)
        if extended_clause.else_clause:
            cextended_clause = convert_extended_clause(extended_clause.else_clause)
            return["elif_clause", crepr_condition, cstatements, cextended_clause ]
        return ["elif_clause", crepr_condition, cstatements ]

    if extended_clause.tag == "else_clause":
        print "WORKING THROUGH ELSE:", extended_clause
        cstatements = convert_statements(extended_clause.block)
        return ["else_clause", cstatements ]

    print "NOT ELIF!"
    print "NOT ELSE!"
    print extended_clause
    raise ValueError("Extended clause isn't a else or elif clause: %s" % repr(extended_clause))


def convert_if_statement(if_statement):
    print "WORKING THROUGH IF:", if_statement
    crepr_condition = convert_arg(if_statement.condition)
    cstatements = convert_statements(if_statement.block)
    if if_statement.else_clause:
        cextended_clause = convert_extended_clause(if_statement.else_clause)
        return["if_statement", crepr_condition, cstatements, cextended_clause ]
    return ["if_statement", crepr_condition, cstatements ]

def convert_break_statement(break_statement):
    return ["break_statement"]

def convert_continue_statement(continue_statement):
    return ["continue_statement"]

def convert_expression_statement(statement):
    print "CONVERTING EXPRESSION STATEMENTS", statement.value
    print "EXPRESSION STATEMENT", statement.value.tag
    crepr = convert_arg(statement.value)
    print "RECHED HERE"
    print "CONVERTED ", crepr
    return ["expression_statement", crepr]


def convert_statements(AST):
    cstatements = []
    statements = AST.statements
    for statement in statements:
        try:
            if statement.tag == "assignment_statement":
                cstatement = convert_assignment(statement)
                print cstatement
                cstatements.append(cstatement)
            elif statement.tag == "print_statement":
                cstatement = convert_print(statement)
                cstatements.append(cstatement)
            elif statement.tag == "expression_statement":
                cstatement = convert_expression_statement(statement)
                cstatements.append(cstatement)
            elif statement.tag == "while_statement":
                cstatement = convert_while_statement(statement)
                cstatements.append(cstatement)
            elif statement.tag == "for_statement":
                cstatement = convert_for_statement(statement)
                cstatements.append(cstatement)
            elif statement.tag == "if_statement":
                cstatement = convert_if_statement(statement)
                cstatements.append(cstatement)
            elif statement.tag == "break_statement":
                cstatement = convert_break_statement(statement)
                cstatements.append(cstatement)
            elif statement.tag == "continue_statement":
                cstatement = convert_continue_statement(statement)
                cstatements.append(cstatement)
            else:
                print "SKIPPING STATEMENT", statement.tag

        except CannotConvert:
            print "REACHED HERE FOR", statement
            pass
    return cstatements

def ast_to_cst(program_name, AST):
    cst = {}

    ast_includes = [ x.replace("#include ","") for x in AST.includes ]
    print "AST.includes = ", ast_includes
    # Extract and handle variables
    pvariables = find_variables(AST)
    cvariables = []
    ctypes = {}
    # includes = []
    includes = [ x.replace("#include ","") for x in AST.includes ]
    names = pvariables.keys()
    names.sort()
    for name in names:
        ctype = python_type_to_c_type(pvariables[name])
        identifier = [ "identifier", ctype, name ]
        cvariables.append(identifier)
        ctypes[ctype] = True

    print "cvariables",cvariables

    cstatements = convert_statements(AST)
    print cstatements

    # Based on variables, update includes
    for ctype in ctypes:
        inc = includes_for_ctype(ctype)
        if inc:
            includes.append(inc)

    print "INCLUDES::", includes
    # Based on statements, update includes
    for cstatement in cstatements:
        inc = includes_for_cstatement(cstatement)
        if inc:
            includes.append(inc)

    program = {}
    program["name"] = program_name
    program["includes"] = sorted(includes)
    program["main"] = {}
    program["main"]["c_frame"] = {}
    program["main"]["c_frame"]["identifiers"] = cvariables
    program["main"]["c_frame"]["statements"] = cstatements

    return { "PROGRAM" : program }


if __name__ == "__main__":
    AST =   ['program',
             ['statements',
              [['assignment_statement',
                ['first', 'IDENTIFIER', 1],
                ['ASSIGN', '='],
                ['value_literal', 1, 'NUMBER', 'INT', 1]],
               ['assignment_statement',
                ['second', 'IDENTIFIER', 2],
                ['ASSIGN', '='],
                ['value_literal', 2, 'NUMBER', 'INT', 2]],
               ['assignment_statement',
                ['third', 'IDENTIFIER', 3],
                ['ASSIGN', '='],
                ['value_literal', 3, 'NUMBER', 'INT', 3]],
               ['print_statement',
                [['value_literal', 'first', 'IDENTIFIER', 5],
                 ['value_literal', 'second', 'IDENTIFIER', 5],
                 ['value_literal', 'third', 'IDENTIFIER', 5]]],
               ['print_statement',
                [['value_literal', 1, 'NUMBER', 'INT', 6],
                 ['value_literal', 2, 'NUMBER', 'INT', 6],
                 ['value_literal', 'hello', 'STRING', 6]]]]]]

    expect = {
        'PROGRAM': {'includes': ['<iostream>', '<iostream>'],
             'main': {'c_frame': {'identifiers': [['identifier', 'int', 'second'],
                                                  ['identifier', 'int', 'third'],
                                                  ['identifier', 'int', 'first']],
                                  'statements': [['assignment', 'first', '=', '1'],
                                                 ['assignment', 'second', '=', '2'],
                                                 ['assignment', 'third', '=', '3'],
                                                 ['print_statement',
                                                      ['identifier', 'first'],
                                                      ['identifier', 'second'],
                                                      ['identifier', 'third']],
                                                 ['print_statement',
                                                      ["integer", 1],
                                                      ["integer", 2],
                                                      ["string", 'hello']]]}},
             'name': 'hello_world_mixed'}}

    AST = ['program',
           ['statements',
            [['print_statement',
              [['operator_function',
                'plus',
                ['value_literal', 1, 'NUMBER', 'INT', 1],
                ['value_literal', 1, 'NUMBER', 'INT', 1]]]]]]]

    expect = {
        'PROGRAM': {'includes': ['<iostream>'],
             'main': {'c_frame': {'identifiers': [],
                                  'statements': [['print_statement', ['op', 'plus', 
                                                                            ["integer", 1],
                                                                            ["integer", 1]]]
,
                                                ] }},
             'name': 'hello_operators'}}


    actual = ast_to_cst("hello_operators", AST)

    print "actual == expect --->", actual == expect

    import json
    import pprint
    # print json.dumps(actual, indent=4)
    pprint.pprint(actual)
    pprint.pprint(expect)




#!/usr/bin/python

files = {'iterators.cpp': '#include "iterators.hpp"\n\nstruct range : public Generator<int> {\n    int maxCount;\n    int index;\n    range() :maxCount(0), index(0) {     };\n    range(int max) :maxCount(max), index(0) {     };\n    ~range() {     };\n\n    virtual int next() {\n    GENERATOR_CODE_START\n    while (index<maxCount) {\n         YIELD(index);\n         index += 1;\n    }\n    GENERATOR_CODE_END\n    }\n};\n', 'iterators.hpp': '#ifndef PYXIE_ITERATORS_HPP\n#define PYXIE_ITERATORS_HPP\n\n/*\n * Python Style Iterators in C++, based in part on experiment work in Kamaelia\n * \n */\n\n#include <exception>\n\n#define GENERATOR_CODE_START  if (this->__generator_state == -1)      \\\n                     { throw StopIteration(); }         \\\n                     switch(this->__generator_state)         \\\n                     { default:\n#define YIELD(v)         this->__generator_state = __LINE__; \\\n                         return (v);                         \\\n                         case __LINE__:\n#define GENERATOR_CODE_END     };                                      \\\n                     this->__generator_state = -1; throw StopIteration();\n\nclass StopIteration : public std::exception {\n  virtual const char* what() const throw()\n  {\n    return "StopIteration";\n  }\n};\n\ntemplate<class T>\nstruct Iterator {\n    virtual T next() {  }\n};\n\ntemplate<class T>\nclass Generator : public Iterator<T> {\n  protected:\n    int __generator_state;\n  public:\n    Generator()  {     };\n    ~Generator() {     };\n};\n\n#endif\n', 'iterators_test.cpp': '/*\n * Test of a C++ version of Python style generators.\n * \n */\n\n#include <iostream>\n#include "iterators.cpp"\n/*\n\nWhile it may be a little unclear, this C++ program is equivalent\nto this python program:\n\nfor count in range(5):\n    print count,\n\nprint\n\nIt doesn\'t do this:\n\nfor(int count=0; count<5; count++) {\n    std::cout << count;\n}\nstd::cout << "." << std::endl;\n\nBecause doing it the way we do makes it avoid treating for/range as a\nspecial case, meaning we solve the harder problem first.  Optimisations\nspecific to certain code structures can come later.\n\n*/\n\n\nint main(int argc, char* argv[]) {\n    // Code to be replaced with a range() style iterator\n    int count;\n    range range_gen = range(5);\n    while (true) {\n        try {\n            count = range_gen.next();\n        } catch (StopIteration s) {\n            break;\n        }\n        std::cout << count;\n    }\n    std::cout << "." << std::endl;\n\n    return 0;\n}\n'}




#!/usr/bin/python

"""
This file contains aspects related to modelling functions.

Python functions return values, these values can be:
* None
* Basic objects - int, float, string, boolean
* Complex objects - lists, dicts, instances from user classes
* Iterators/Generators

Functions can be:
* Built in
* User defined.

This file is primarily focussed on built in functions at the
moment, to facilitate "range"
"""

# built_ins = ("range", "print")

# Might be worth considering bunging all the builtins into seperate modules
# to be auto-discovered to populate this structure. It'd make creating custom
# versions easier potentially.

# This won't actually be used, it's here to help figure out what's needed
range_raw_code_template = """
    int count;
    range range_gen = range(5);
    while (true) {
        try {
            count = range_gen.next();
        } catch (StopIteration s) {
            break;
        }
        std::cout << count;
    }
"""


# This might well be used
iterated_code_variable_declarations_template = """
    %(ITERATED_TYPE)s %(IDENTIFIER)s;
    %(ITERATOR_TYPE)s %(ITERATOR)s_gen_%(UNIQIFIER)s = %(ITERATOR)s;
"""

# This might well be used
iterated_code_actual_code_template = """
    while (true) {
        try {
            %(IDENTIFIER)s = %(ITERATOR)s_gen_%(UNIQIFIER)s.next();
        } catch (StopIteration s) {
            break;
        }
        %(BODY)s # Itself uses %(IDENTIFIER)s
    }
"""

# This next line doesn't seem likely to be what we want here.
iterated_code_template = iterated_code_variable_declarations_template + iterated_code_actual_code_template
"""
    %(ITERATED_TYPE)s %(IDENTIFIER)s;
    %(ITERATOR)s %(ITERATOR)s_gen_%(UNIQIFIER)s = %(ITERATED_TYPE)s %(ITERATOR)s(5);
    while (true) {
        try {
            %(IDENTIFIER)s = %(ITERATOR)s_gen_%(UNIQIFIER)s.next();
        } catch (StopIteration s) {
            break;
        }
        %(BODY)s # Itself uses %(IDENTIFIER)s
    }
"""

in_the_above = {
        "ITERATED_TYPE" : "The type of the values created by this iterator",
        "IDENTIFIER" : "The identifier that takes the value created by the iterator for use in the loop body",
        "ITERATOR" : "The specific iterator that gets 'called' creating instances to iterate over",
        "UNIQIFIER" : "A unique id added to ensure that instances in the code as used are all new",
        "BODY" : "A set of 1 or more instructions that uses the values",
    }



builtins = {
             "range": {
                         "iterator": True,
                         "values_type": "integer",
                         "iterator_ctype" : "range",  # The name of the C function/struct/class
                         "code_template" : iterated_code_template, # Probably not what we want
                         "variable_declarations_template" :  iterated_code_variable_declarations_template,
                         "actual_code_template" :  iterated_code_actual_code_template,
                     }
            }


templates = { "iterator" : { "declarations" : iterated_code_variable_declarations_template,
                             "code" : iterated_code_actual_code_template,
                           }
            }



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from distutils.core import setup
from distutils.version import LooseVersion
import os

def is_package(path):
    return (
        os.path.isdir(path) and
        os.path.isfile(os.path.join(path, '__init__.py'))
        )

def find_packages(path, base="" ):
    """ Find all packages in path """
    packages = {}
    for item in os.listdir(path):
        dir = os.path.join(path, item)
        if os.path.isdir(dir):
            if base:
                module_name = "%(base)s.%(item)s" % vars()
            else:
                module_name = item
            if is_package( dir ):
                packages[module_name] = dir
            packages.update(find_packages(dir, module_name))
    return packages


packages = find_packages(".")
package_names = packages.keys()

setup(name = "pyxie",
      version = "0.0.15",
      description = "Little Python to C++ Compiler",
      url='http://www.sparkslabs.com/pyxie/',
      author='Michael Sparks (sparkslabs)',
      author_email="sparks.m@gmail.com",
      license='Apache Software License',

      scripts = [
                  'bin/pyxie',
                ],

      packages = package_names,
      package_dir = packages,
      package_data={},
      long_description = """
Pyxie -- A Little Python to C++ Compiler
========================================

### What job will this do?

The aim of this project is to allow a user to write code in a familiar high level
language that can then be compiled to run on an arbitrary embedded system - that is
devices with very low power CPUs and very little memory. (ie devices too small to
host a python interpreter/runtime) In particular, it aims to be useful in supporting
The Scout Association's "Digital Maker" badge, but that's some way off!

It's pre-alpha at the moment.

### What does it do?

Currently:

- Recognise simple sequential python programs with simple statements
- Can handle basic conditionals and while loops
- Parse those to an AST
- Can represent equivalent C programs using a concrete C representation (CST)
- Can translate the AST to the CST and then generate C++ code from the CST

That means it can compile one very very simple type of python program
into an equivalent (simple) C++ program, and then compile that.

Example program:

    age = 10
    new_age = 10 +1
    new_age_too = age + 1
    new_age_three = age + new_age_too
    foo = "Hello"
    bar = "World"
    foobar = foo + bar

    print 10-1-2,7
    print 1+2*3*4-5/7,25
    print age, new_age, new_age_too
    print foo, bar, foobar

Example results:

    #include <iostream>
    #include <string>

    using namespace std;

    int main(int argc, char *argv[])
    {
        int age;
        string bar;
        string foo;
        string foobar;
        int new_age;
        int new_age_three;
        int new_age_too;

        age = 10;
        new_age = (10+1);
        new_age_too = (age+1);
        new_age_three = (age+new_age_too);
        foo = "Hello";
        bar = "World";
        foobar = (foo+bar);
        cout << ((10-1)-2) << " " << 7 << endl;
        cout << ((1+((2*3)*4))-(5/7)) << " " << 25 << endl;
        cout << age << " " << new_age << " " << new_age_too << endl;
        cout << foo << " " << bar << " " << foobar << endl;
        return 0;
    }

Additionally, while loops, and comparison operators are supported, allowing
very basic programs to start being useful since you can have sequence, selection
and iteration now. Additionally, function calls are supported from libraries
that we link with.

It's a starting point, not the end point. For that, take a look at the language spec.

Pyxie is intended to be a simple Python to C++ compiler, with a target of
compiling python code such that it can run on a microcontroller - like
Arduino, MSP430 or ARM mbed type devices.

The name is a play on words. Specifically, Python to C++ - can be py2cc or
pycc.  If you try pronouncing "pycc" it can be "pic", "py cc" or pyc-c".
The final one leads to Pixie.

This is unlikely to ever be a completely general python to C++ compiler - if
you're after than look at Shed Skin, or things like Cython, Pyrex, and PyPy.
(in terms of diminishing similarity) The difference this project has from
those is that this project assumes a very small target device.  Something
along the lines of an Atmega 8A, Atmega 328 or more capable.

This is also a difference from MicroPython - which is designed to run on
microcontrollers larger than the Atmega 8A.

In the past I've written a test driven compiler suite, so I'll be following
the same approach here.  It did consider actually making Pyxie use that as a
frontend, but for the moment, I'd like python compatibility.

## Status

For the impatient: this probably does **NOT** do what you want, **yet**.
Check back in a couple of months time :-)

It IS getting there however, and feedback, usecases, devices very welcome.

Specifically:

* Supports variables, sequence, and assignment
* while loops controlled by expressions, possibly involving variables
* while loops can contain break/continue which allows "if" style functionality
* Also have basic conditional operators like "==", "!=", etc.
* This means we can almost start writing useful programs, but in particular can start creating simplistic benchmarks for measuring run speed.

## Influences

Many moons ago, I made a generic language parser which I called SWP (semantic
 whitespace parser), or Gloop.

* https://github.com/sparkslabs/minisnips/tree/master/SWP
* http://www.slideshare.net/kamaelian/swp-a-generic-language-parser

It was an experiment to see if you could write a parser that had no keywords,
or similar, in a completely test driven fashion. ie a bit like a parser for a
Lisp like language that would look like python or ruby. It turns out that you
can and there's lots of interesting things that arise if you do. (Best seen
in the slideshare link)

## Which version of Python?

It's not a complete subset of any particular python, but it's based around the
intersection points in python 2 and 3.  It will be, by definition, a non-dynamic
subset - at least at first.

* For detail as to what's planned for the language, take a look at the language spec.
* For an overview as to the guiding principles, please take a look at project status
* For detail as to what's actually implemented, take a look at language status

These are all a WIP, but becoming more solid.

## Why write this?

Personally, having built something simpler in the past, I know I'd find it
useful. (I use python rather than C++ often because I can write more quicker
with the former). Also, I work with kids in my spare time, and it opens up
options there.

I've written something like this for work last year, but that was much more
limited and restricted in both aspiration and implementation. This rewrite is
something I've done on my own time, with my own tools, from scratch, which
allows me to share this with others.

Major changes:

* This aims to be a more rounded implementation
* This performs transforms from an AST (abstract syntax tree) to a CCR (concrete
  code representation), rather than munging code directly from a concrete parse
  tree.

That potentially allows other things, like creation of visual representations
of programs from code as well.

## Is this part of any larger project?

No. It could be used by others, but it's got a definite goal - to allow the
use of a "little" python to program devices which are too small to host a python
runtime.

If anything, it's a continuation of the personal itch around SWP (mentioned above)
from about 10 years ago. Unlike that though, it's much, much better structured.

One thing that may happen though is the ability to take python classes and
derive iotoy device implementations/interfaces directly. (since iotoy was
inspired heavily by python introspection) That's quite some time off.

## Release History

Release History: (and highlights)

* 0.0.15 - 2015-07-18 - clib converted to py clib for adding to build directory
* 0.0.14 - 2015-07-18 - For loops implemented. Added clib code, C++ generator implementation, FOR loop style test harness, parsing and basic analysis of of FOR loops using a range interator
* 0.0.13 - 2015-06-21 - if/elif/else,conditionals/boolean/parenthesised expressions.
* 0.0.12 - 2015-06-16 - While loops, break/continue, Website, comparison operators, simple benchmark test
* 0.0.11 - 2015-06-06 - Function calls; inclusion of custom  C++ headers; empty statements; language spec updates
* 0.0.10 - 2015-06-03 - Analysis phase to make type inference work better. Lots of related changes. Implementation of expression statements.
* 0.0.9 - 2015-05-23 - Grammar changed to be left, not right recursive. (Fixes precedence in un-bracketed expressions) Added standalone compilation mode - outputs binaries from python code.
* 0.0.8 - 2015-05-13 - Internally switch over to using node objects for structure - resulting in better parsing of expressions with variables and better type inference.
* 0.0.7 - 2015-04-29 - Structural, testing improvements, infix operators expressions (+ - * / ) for integers, precdence fixes
* 0.0.6 - 2015-04-26 - Character Literals, "plus" expressions, build/test improvements
* 0.0.5 - 2015-04-23 - Core lexical analysis now matches language spec, including blocks
* 0.0.4 - 2015-04-22 - Mixed literals in print statements
* 0.0.3 - 2015-04-21 - Ability to print & work with a small number of variables
* 0.0.2 - 2015-03-30 - supports basic assignment
* 0.0.1 - Unreleased - rolled into 0.0.2 - Initial structure

Michael Sparks, June 2015
"""
      )



#!/usr/bin/python
#
# Copyright 2015 Michael Sparks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pyxie.model.functions
import pprint
from pyxie.model.functions import builtins


blank_line = ""
unique_id = 0
source = []
def Print(*args):
    y = " ".join([str(x) for x in args])
    source.append(y)

def reset_parser():
    global source
    source = []

def mkStatement(statement_spec):
    ss = statement_spec
    if ss[0] == "assignment":
        return Assigment( ss[1], ss[3], ss[2])

    elif ss[0] == "print_statement":
        return PrintStatement(*statement_spec[1:])

    elif ss[0] == "expression_statement":
        return ExpressionStatement(*statement_spec[1:])

    elif ss[0] == "while_statement":
        return WhileStatement(*statement_spec[1:])

    elif ss[0] == "for_statement":
        return ForStatement(*statement_spec[1:])

    elif ss[0] == "if_statement":
        if len(ss) == 3:
            return IfStatement(ss[1], ss[2])
        if len(ss) == 4:
            return IfStatement(ss[1], ss[2], ss[3])
        raise NotImplementedError("Have not yet implemented else clauses...")

    elif ss[0] == "break_statement":
        return BreakStatement()

    elif ss[0] == "continue_statement":
        return ContinueStatement()

    else:
        print "Unknown statement type", ss[0], ss

class C_Program(object):
    def __init__(self):
        self.includes = []
        self.main_cframe = C_Frame()
        self.name = ""

    @classmethod
    def fromjson(klass, json):
        program = klass()
        prog_desc = json["PROGRAM"]
        program.name = prog_desc["name"]
        program.includes = list(prog_desc["includes"])
        main_spec = prog_desc["main"]["c_frame"]
        for identifier in main_spec["identifiers"]:
            program.main_cframe.identifiers.append(Identifier(identifier[1], identifier[2]))

        for statement in main_spec["statements"]:
            conc_statement = mkStatement(statement)
            program.main_cframe.statements.append(conc_statement)


        return program

    def generate(self):
        frame_lines = self.main_cframe.concrete()
        seen = {}
        for include in self.includes:
            if not seen.get(include, False):
                # Only output each include once
                Print( "#include "+ include )
                seen[include] = True

        Print()
        Print("using namespace std;")
        print_def = """
#include <iostream>
#include "iterators.cpp"
template<typename T>
void Print(T x) {
    std::cout << x;
}
"""
        Print(print_def)
        Print()
        Print("int main(int argc, char *argv[])")
        Print("{")
        for line in frame_lines:
            Print("    "+ line)
        Print("    "+ "return 0;")
        Print("}")

    def json(self):
        return { "PROGRAM": {"name": self.name,
                             "includes" : self.includes,
                             "main": self.main_cframe.json()}
               }

class C_Frame(object):
    def __init__(self):
        self.identifiers = []
        self.statements = []

    def json(self):
        for y in self.statements:
            Print(y)
        return {"c_frame": {"identifiers" : [ x.json() for x in self.identifiers ],
                            "statements" : [y.json() for y in self.statements ] }
               }

    def concrete(self):
        block = []
        for identifier in self.identifiers:
            decl_code = identifier.decl_code()
            block.append(decl_code + ";")

        block.append(blank_line)
        for statement in self.statements:
            if statement:
                code = statement.code()
                if code is None:
                    print "STATEMENT IS None, WHY?", statement
                block.append(code + ";")
        return block

class Identifier(object):
    def __init__(self, ctype, name):
        self.ctype = ctype
        self.name = name

    def json(self):
        return ["identifier", self.ctype, self.name ]

    def decl_code(self):
        return self.ctype + " " + self.name

class Assigment(object):
    def __init__(self, lvalue, rvalue, assigntype="="):
        self.lvalue = lvalue
        self.rvalue = rvalue
        self.assigntype = assigntype

    def json(self):
        return ["assignment", self.lvalue, self.assigntype, self.rvalue ]

    def code(self):
        print self.rvalue
        if type(self.rvalue) == list:
            print "Aha!",self.rvalue
            crvalue = ArgumentList(self.rvalue)
            crvalue = crvalue.code()
            print "AHA!", crvalue
        else:
            crvalue = self.rvalue
        return self.lvalue + " "+self.assigntype+" " + crvalue


class BreakStatement(object):
    def json(self):
        return ["break_statement"]

    def code(self):
        return "break"

class ContinueStatement(object):
    def json(self):
        return ["continue_statement"]

    def code(self):
        return "continue"

class ExpressionStatement(object):
    def __init__(self, expression):
        self.expression = expression

    def json(self):
        return ["expression_statement", self.expression ]

    def code(self):
        print self.expression
        if type(self.expression) == list:
            cvalue = ArgumentList(self.expression)
            cvalue = cvalue.code()
        else:
            raise Exception("Fail to understand how to generate code for %s" % repr(self.expression) )

        return cvalue

def todo(*args):
    print("TODO", " ".join([repr(x) for x in args]))

class ArgumentList(object):
    def __init__(self, *args):
        self.args = args
    def json(self):
        return list(self.args[:])

    def code_op_F(self, arg):
        if arg[1] == "plus": return "+"
        if arg[1] == "minus": return "-"
        if arg[1] == "times": return "*"
        if arg[1] == "divide": return "/"
        if arg[1] == "boolean_or": return " || "
        if arg[1] == "boolean_and": return " && "
        if arg[1] == "boolean_not": return " ! "

        if arg[1] in ["<", ">", "==", ">=", "<=", "!="]:
            return arg[1]

        if arg[1] == "<>": return "!="

        return None

    def code_op(self,arg):
        c_op = self.code_op_F(arg)
        if c_op:

            if len(arg) == 4:
                arg1 = arg[2]
                arg2 = arg[3]
                # We would like to try to assert here that the values on both sides
                # are integers, but at present we cannot do that, since we use too simplistic
                # a structure. If I remove that constraint, we can generate more code.
                # But we will need to revisit this.
                lit_arg1 = self.code_arg(arg1)
                lit_arg2 = self.code_arg(arg2)

                result = "(" + lit_arg1 + c_op + lit_arg2 + ")"
                return result
            if len(arg) == 3:
                arg1 = arg[2]
                # We would like to try to assert here that the values on both sides
                # are integers, but at present we cannot do that, since we use too simplistic
                # a structure. If I remove that constraint, we can generate more code.
                # But we will need to revisit this.
                lit_arg1 = self.code_arg(arg1)

                result = "(" + c_op + lit_arg1 + ")"
                return result

        todo("Handle code ops for anything other than plus/int,int")
        raise NotImplementedError("Handle code ops for anything other than plus/int,int" + repr(arg))


    def code_arg(self, arg):
        if arg[0] == "identifier":
            return arg[1]
        elif arg[0] == "integer":
            r = repr(arg[1])
            if arg[1]<0:
                r = "(" + r + ")"
            return r
        elif arg[0] == "string":
            carg = arg[1].replace('"', '\\"')
            return '"' + carg + '"' # Force double quotes
        elif arg[0] == "double":
            return repr(arg[1])
        elif arg[0] == "boolean":
            return arg[1]
        elif arg[0] == "op":
            return self.code_op(arg)
        elif arg[0] == "function_call":
            code_gen = FunctionCall(arg[1],arg[2])
            return code_gen.code()
            print "We don't know how to generate code for function calls yet", arg
            return ""
        todo("Handle print value types that are more than the basic types")
        raise NotImplementedError("Handle print value types that are more than the basic types" + repr(arg))

    def code_list(self):
        cargs = []
        for arg in self.args:
            c_str = self.code_arg(arg)
            cargs.append(c_str)
        return cargs

    def code(self):
        return ",".join(self.code_list())

class FunctionCall(object):
    def __init__(self, identifier, args):
        self.args = args
        self.identifier = identifier
        self.arg_list = ArgumentList(*args)

    def json(self):
        return ["function_call", self.identifier ] + self.arg_list.json()

    def code(self):
        return self.identifier[1] + "(" + ", ".join(self.arg_list.code_list())+ ")"
#         return "cout << " + " << \" \" << ".join(self.arg_list.code_list()) + " << endl"


class PrintStatement(object):
    def __init__(self, *args):
        self.args = args
        self.arg_list = ArgumentList(*args)

    def json(self):
        return ["print_statement" ] + self.arg_list.json()

    def code(self):
        return "cout << " + " << \" \" << ".join(self.arg_list.code_list()) + " << endl"


class WhileStatement(object):
    def __init__(self, condition, *statements):
        self.raw_condition = condition
        self.raw_statements = list(statements)
        self.block_cframe = C_Frame()

        for statement in self.raw_statements:
            conc_statement = mkStatement(statement)
            self.block_cframe.statements.append(conc_statement)

    def json(self):
        return ["while_statement", self.raw_condition] + self.raw_statements

    def code(self):
        code = "while"
        code += "("
        code += ArgumentList(self.raw_condition).code()
        code += ") {"
        code += "\n".join( self.block_cframe.concrete() )
        code += "}"
        return code

class ForStatement(object):
    def __init__(self, identifier, iterable, statements, for_statement_pynode):
        self.raw_identifier = identifier
        self.raw_iterable = iterable
        self.raw_statements = list(statements)
        self.block_cframe = C_Frame()
        self.for_statement_pynode = for_statement_pynode

        for statement in self.raw_statements:
            conc_statement = mkStatement(statement)
            self.block_cframe.statements.append(conc_statement)

    def json(self):
        return ["for_statement", self.raw_condition, self.raw_iterable, self.raw_statements ]

    def code(self):
        #
        # The following is the sort of code we want to generate inline.
        # This actually implements what we're after.
        # However it requires some declarations to be in place really of the the find variables kind
        # That said, C++ allows inline delcarations like this (for good or ill), so maybe just do
        # this in the first instance, and leave improvements for refactoring???
        #
        # Actually this discussion is useful, but it turns out to allow nested generator usage
        # ie for x in range(5): for y in range(x): print x,y
        # It requires this to be done closer to what could be viewed as "properly"
        #
#            %(ITERATOR_TYPE)s %(ITERATOR_NAME)s = %(ITERATOR_EXPRESSION)s;
        template = """
            %(ITERATOR_NAME)s = %(ITERATOR_EXPRESSION)s;
            while (true) {
                try {
                    %(IDENTIFIER)s = %(ITERATOR_NAME)s.next();
                } catch (StopIteration s) {
                    break;
                }
                %(BODY)s // Itself uses %(IDENTIFIER)s
            }
        """
        global unique_id
        unique_id = unique_id+1

        iterator_ctype = "range"
        iterator_expression = self.raw_iterable ## NOTE THIS RESULTS IN ['iterate_over', ['function_call', ['identifier', 'range'], [['integer', 5]]]]- so that needs work

        print "OK, USE THIS THEN", self.for_statement_pynode.expression.ivalue_name
        iterator_name = self.for_statement_pynode.expression.ivalue_name

        if self.raw_iterable[0] == "iterator":
            iterable = self.raw_iterable[1]
            if iterable[0] == "function_call":
                assert iterable[1][0] == "identifier"
                identifier = iterable[1][1]
                if identifier in builtins:
                    print "YAY"
                    pprint.pprint ( builtins[identifier] )
                    iterator_ctype = builtins[identifier]["iterator_ctype"]
                    print "iterator_name -- ", iterator_ctype
                    print "YAY"

                print iterable
                fcall = FunctionCall(iterable[1],iterable[2])
                fcall_code = fcall.code()
                iterator_expression = fcall_code

        template_args =  { "ITERATOR_TYPE": iterator_ctype,
                           "ITERATOR_EXPRESSION": iterator_expression,
                           "ITERATOR_NAME": iterator_name,
                           "UNIQIFIER":repr(unique_id),
                           "IDENTIFIER": self.raw_identifier,
                           "BODY": "\n".join( self.block_cframe.concrete() )
                         }



        print "=== template args ================================================="
        pprint.pprint (template_args )
        print "--- template args -------------------------------------------------"
        if self.for_statement_pynode.expression.isiterator:
            identifier = self.for_statement_pynode.identifier
            print identifier.get_type()
        print "=== end extractable template args ================================="
        print template % template_args

        print "----------------------------------------------------------"
        print dir(self.for_statement_pynode)
        print repr(self.for_statement_pynode)
        print "----------------------------------------------------------"
#        raise NotImplementedError("Haven't finished implementing ForStatement yet...")
        code = template % template_args

        print "CODE", code

        return code

class IfStatement(object):
    def __init__(self, condition, statements, extended_clause=None):
        self.raw_condition = condition
        self.raw_statements = list(statements)
        self.extended_clause = extended_clause

        self.block_cframe = C_Frame()
        for statement in self.raw_statements:
            conc_statement = mkStatement(statement)
            self.block_cframe.statements.append(conc_statement)

    def json(self):
        result = ["if_statement", self.raw_condition, self.raw_statements ]
        if self.extended_clause is not None:
            result.append(extended_clauses.json())

    def code(self):
        extended_clauses_code = None
        condition_code = ArgumentList(self.raw_condition).code()
        block_code = "\n".join( self.block_cframe.concrete() )
        if self.extended_clause:
            if self.extended_clause[0] == "elif_clause":
                print self.extended_clause[0]
                condition = self.extended_clause[1]
                statements =  self.extended_clause[2]
                if len(self.extended_clause) == 4:
                    extended_sub_clause =  self.extended_clause[3]
                    extended_clauses_code = ElIfClause( condition, statements, extended_sub_clause ).code()
                else:
                    extended_clauses_code = ElIfClause( condition, statements ).code()

            if self.extended_clause[0] == "else_clause":
                print "***************************************************"
                print self.extended_clause[0]
                statements =  self.extended_clause[1]
                extended_clauses_code = ElseClause( statements ).code()

        code = "if ( %s ) { %s } " % (condition_code, block_code )
        if extended_clauses_code:
            code = code + extended_clauses_code
        return code

class ElIfClause(object):
    def __init__(self, condition, statements, extended_clause=None):
        self.raw_condition = condition
        self.raw_statements = list(statements)
        self.extended_clause = extended_clause

        self.block_cframe = C_Frame()
        for statement in self.raw_statements:
            conc_statement = mkStatement(statement)
            self.block_cframe.statements.append(conc_statement)

    def json(self):
        result = ["elif_clause", self.raw_condition, self.raw_statements ]
        if self.extended_clause is not None:
            result.append(extended_clause.json())

    def code(self):
        extended_clauses_code = None
        condition_code = ArgumentList(self.raw_condition).code()
        block_code = "\n".join( self.block_cframe.concrete() )
        if self.extended_clause:
            if self.extended_clause[0] == "elif_clause":
                print "***************************************************"
                print self.extended_clause[0]
                condition = self.extended_clause[1]
                statements =  self.extended_clause[2]

                if len(self.extended_clause) == 4:
                    extended_sub_clause =  self.extended_clause[3]
                    extended_clauses_code = ElIfClause( condition, statements, extended_sub_clause ).code()
                else:
                    extended_clauses_code = ElIfClause( condition, statements ).code()


            if self.extended_clause[0] == "else_clause":
                print "***************************************************"
                print self.extended_clause[0]
                statements =  self.extended_clause[1]
                extended_clauses_code = ElseClause( statements ).code()

        code = "else if ( %s ) { %s } " % (condition_code, block_code )
        if extended_clauses_code:
            code = code + extended_clauses_code
        return code


class ElseClause(object):
    def __init__(self, statements):
        self.raw_statements = list(statements)

        self.block_cframe = C_Frame()
        for statement in self.raw_statements:
            conc_statement = mkStatement(statement)
            self.block_cframe.statements.append(conc_statement)

    def json(self):
        result = ["elif_clause", self.raw_statements ]

    def code(self):
        extended_clauses_code = None
        block_code = "\n".join( self.block_cframe.concrete() )

        code = "else { %s } " % (block_code, )

        return code

makefile_tmpl = """
all :
	g++ %(filename)s.c -o %(filename)s
"""

def build_program(json):
    json = C_Program.fromjson(json)


if __name__ == "__main__":



    if 1:
        progj = {'PROGRAM': {'includes': ['<iostream>'],
                             'main': {'c_frame': {'identifiers': [],
                                                  'statements': [['print_statement',
                                                                  ['op',
                                                                   'plus',
                                                                   ['integer', 1],
                                                                   ['integer', 1]]]]}},
                             'name': 'hello_operators'}}

        program = C_Program.fromjson(progj)
        print program

        program.generate()
        import time
        import pprint
        import os

        pprint.pprint(source, width=200)
        now = int(time.time())
        dirname = str(now - 1427000000)
        # Print("BUILDING PROGRAM", dirname)
        os.mkdir(dirname)
        f = open(os.path.join(dirname,program.name+".c"), "w")
        for line in source:
            f.write(line)
            f.write("\n")
        f.close()

        makefile = makefile_tmpl % {"filename": program.name }
        f = open(os.path.join(dirname,"Makefile"), "w")
        f.write(makefile)
        f.close()

        os.chdir(dirname)
        os.system("make")



    # Build an example concrete syntax tree
    if 0:
        program = C_Program()

        program.name = ("hello")
        program.includes.append("<iostream>")
        program.includes.append("<string>")

        main = program.main_cframe
        main.identifiers.append(Identifier("string", "greeting"))
        main.identifiers.append(Identifier("string", "name"))
        main.statements.append(Assigment("greeting", '"hello"'))
        main.statements.append(Assigment("name", '"world"'))
        main.statements.append(PrintStatement("greeting", "name"))

        import pprint

        progj = program.json()

        pprint.pprint(progj)

        program_clone = C_Program.fromjson(progj)
        progj2 = program_clone.json()

        program = program_clone

        Print("--------------------------------------------------------------")
        pprint.pprint(progj2)

        program.generate()

        import time
        import pprint
        import os

        pprint.pprint(source, width=200)
        now = int(time.time())
        dirname = str(now - 1427000000)
        Print("BUILDING PROGRAM", dirname)
        os.mkdir(dirname)
        f = open(os.path.join(dirname,program.name+".c"), "w")
        for line in source:
            f.write(line)
            f.write("\n")
        f.close()

        makefile = makefile_tmpl % {"filename": program.name }
        f = open(os.path.join(dirname,"Makefile"), "w")
        f.write(makefile)
        f.close()

        os.chdir(dirname)
        os.system("make")

